import contextvars
import functools
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from typing import (
    Any,
    AsyncContextManager,
    AsyncGenerator,
    Awaitable,
    Callable,
    ContextManager,
    TypeVar,
    Union,
    cast,
)

import anyio

from di._inspect import is_async_gen_callable, is_coroutine_callable, is_gen_callable
from di.dependency import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    DependencyProviderType,
    DependencyType,
    GeneratorProvider,
)

T = TypeVar("T")


def callable_in_thread_pool(call: Callable[..., T]) -> Callable[..., Awaitable[T]]:
    def inner(*args: Any, **kwargs: Any) -> T:
        # Ensure we run in the same context
        child = functools.partial(call, *args, **kwargs)
        context = contextvars.copy_context()
        return cast(Awaitable[T], anyio.to_thread.run_sync(context.run, child))  # type: ignore

    return inner  # type: ignore


def context_manager_in_threadpool(
    call: Callable[..., ContextManager[T]]
) -> Callable[..., AsyncContextManager[T]]:
    @asynccontextmanager
    async def inner(*args: Any, **kwds: Any) -> AsyncGenerator[T, None]:
        cm = call(*args, **kwds)
        try:
            yield await callable_in_thread_pool(cm.__enter__)()
        except Exception as e:
            ok = await callable_in_thread_pool(cm.__exit__)(type(e), e, None)
            if not ok:
                raise e
        else:
            await callable_in_thread_pool(cm.__exit__)(None, None, None)

    return inner


def bind_async_context_manager(
    cm: Callable[..., AsyncContextManager[T]], stack: AsyncExitStack
) -> Callable[..., Awaitable[T]]:
    async def inner(*args: Any, **kwargs: Any):
        return await stack.enter_async_context(cm(*args, **kwargs))

    return inner


def bind_sync_context_manager(
    cm: Callable[..., ContextManager[T]], stack: AsyncExitStack
) -> Callable[..., T]:
    def inner(*args: Any, **kwargs: Any):
        return stack.enter_context(cm(*args, **kwargs))

    return inner


def bind_to_stack_as_awaitable(
    call: DependencyProviderType[DependencyType],
    stack: AsyncExitStack,
) -> Callable[..., Awaitable[DependencyType]]:
    if is_async_gen_callable(call):
        call = cast(AsyncGeneratorProvider[DependencyType], call)
        return bind_async_context_manager(asynccontextmanager(call), stack)
    if is_gen_callable(call):
        call = cast(GeneratorProvider[DependencyType], call)
        return bind_async_context_manager(
            context_manager_in_threadpool(contextmanager(call)), stack
        )
    if not is_coroutine_callable(call):
        call = cast(CallableProvider[DependencyType], call)
        return callable_in_thread_pool(call)
    call = cast(CoroutineProvider[DependencyType], call)
    return call


def bind_to_stack_as_def_callable(
    call: Union[GeneratorProvider[DependencyType], CallableProvider[DependencyType]],
    stack: AsyncExitStack,
) -> Callable[..., DependencyType]:
    if is_gen_callable(call):
        call = cast(GeneratorProvider[DependencyType], call)
        return bind_sync_context_manager(contextmanager(call), stack)
    return cast(Callable[..., T], call)



from collections import deque
from contextlib import asynccontextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from typing import (
    Any,
    AsyncGenerator,
    ContextManager,
    Deque,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    cast,
)

from anyio import create_task_group
from anyio.abc import TaskGroup

from di._concurrency import bind_to_stack_as_awaitable, bind_to_stack_as_def_callable
from di._inspect import DependencyParameter, is_coroutine_callable
from di._state import ContainerState
from di._task import Task
from di._topsort import topsort
from di.dependency import (
    DependantProtocol,
    Dependency,
    DependencyProvider,
    DependencyProviderType,
    DependencyType,
    Scope,
)
from di.exceptions import (
    DependencyRegistryError,
    DuplicateScopeError,
    ScopeViolationError,
    UnknownScopeError,
)

_UNSET_SCOPE: Scope = object()


@dataclass
class SolvedDependency(Generic[DependencyType]):
    """Representation of a fully solved dependency.

    A fully solved dependency consists of:
    - A DAG of sub-dependency paramters.
    - A topologically sorted order of execution, where each sublist represents a
    group of dependencies that can be executed in parallel.
    """

    dependency: DependantProtocol[DependencyType]
    dag: Dict[
        DependantProtocol[Any], Dict[str, DependencyParameter[DependantProtocol[Any]]]
    ]
    topsort: List[List[DependantProtocol[Any]]]


class Container:
    def __init__(self) -> None:
        self._context = ContextVar[ContainerState]("context")
        state = ContainerState()
        # bind ourself so that dependencies can request the container
        self._context.set(state)

    @property
    def _state(self) -> ContainerState:
        return self._context.get()

    @asynccontextmanager
    async def enter_global_scope(self, scope: Scope) -> AsyncGenerator[None, None]:
        """Enter a global scope that is shared amongst threads and coroutines.

        If you enter a global scope in one thread / coroutine, it will propagate to others.
        """
        async with self._state.enter_scope(scope):
            # bind ourself so that dependencies can request the container
            if not self._state.cached_values.contains(Container):
                self._state.cached_values.set(Container, self, scope=scope)
            yield

    @asynccontextmanager
    async def enter_local_scope(self, scope: Scope) -> AsyncGenerator[None, None]:
        """Enter a local scope that is localized to the current thread or coroutine.

        If you enter a global scope in one thread / coroutine, it will NOT propagate to others.
        """
        if scope in self._state.stacks:
            raise DuplicateScopeError(f"Scope {scope} has already been entered!")
        current = self._state
        new = current.copy()
        token = self._context.set(new)
        try:
            async with new.enter_scope(scope):
                # bind ourself so that dependencies can request the container
                if not new.cached_values.contains(Container):
                    new.cached_values.set(Container, self, scope=scope)
                yield
        finally:
            self._context.reset(token)

    def bind(
        self,
        provider: DependantProtocol[DependencyType],
        dependency: DependencyProviderType[DependencyType],
        scope: Optional[Scope] = _UNSET_SCOPE,
    ) -> ContextManager[None]:
        """Bind a new dependency provider for a given dependency.

        This can be used as a function (for a permanent bind, cleared when `scope` is exited)
        or as a context manager (the bind will be cleared when the context manager exits).

        Binds are only identified by the identity of the callable and do not take into account
        the scope or any other data from the dependency they are replacing.

        The `scope` parameter determines the scope for the bind itself.
        The bind will be automatically cleared when that scope is exited.
        If no scope is provided, the current scope is used.
        """
        if scope is _UNSET_SCOPE:
            if len(self._state.scopes) == 0:
                raise ValueError(
                    "If there is no active scope, you MUST supply the `scope` parameter"
                )
            scope = self._state.scopes[-1]  # current scope
        return self._state.bind(provider=provider, dependency=dependency, scope=scope)

    def solve(
        self, dependency: DependantProtocol[DependencyType]
    ) -> SolvedDependency[DependencyType]:
        """Solve a dependency.

        This is done automatically when calling `execute`, but you can store the returned value
        from this function and call `execute_solved` instead if you know that your binds
        will not be changing between calls.
        """

        if self._state.binds.contains(dependency.call):  # type: ignore
            dependency = self._state.binds.get(dependency.call)  # type: ignore

        param_graph: Dict[
            DependantProtocol[Any],
            Dict[str, DependencyParameter[DependantProtocol[Any]]],
        ] = {}

        dep_registry: Dict[DependantProtocol[Any], DependantProtocol[Any]] = {}

        dep_dag: Dict[DependantProtocol[Any], List[DependantProtocol[Any]]] = {}

        def get_params(
            dep: DependantProtocol[Any],
        ) -> Dict[str, DependencyParameter[DependantProtocol[Any]]]:
            params = dep.get_dependencies().copy()
            for keyword, param in params.items():
                assert param.dependency.call is not None
                if self._state.binds.contains(param.dependency.call):
                    params[keyword] = DependencyParameter[Any](
                        dependency=self._state.binds.get(param.dependency.call),
                        parameter=param.parameter,
                    )
            return params

        def check_equivalent(dep: DependantProtocol[Any]):
            if not dep.is_equivalent(dep_registry[dep]):
                raise DependencyRegistryError(
                    f"The dependencies {dep} and {dep_registry[dep]}"
                    " have the same hash but are not equal."
                    " This can be caused by using the same callable / class as a dependency in"
                    " two different scopes, which is usually a mistake"
                    " To work around this, you can subclass or wrap the function so that it"
                    " does not have the same hash/id."
                    " Alternatively, you may provide an implementation of DependencyProtocol"
                    " that uses custom __hash__ and __eq__ semantics."
                )

        q: Deque[DependantProtocol[Any]] = deque([dependency])
        while q:
            dep = q.popleft()
            if dep in dep_registry:
                check_equivalent(dep)
            else:
                dep_registry[dep] = dep
                params = get_params(dep)
                param_graph[dep] = params
                dep_dag[dep] = []
                for param in params.values():
                    subdep = param.dependency
                    dep_dag[dep].append(subdep)
                    if subdep not in dep_registry:
                        q.append(subdep)

        groups = topsort(dependency, dep_dag)
        return SolvedDependency[DependencyType](
            dependency=dependency, dag=param_graph, topsort=groups
        )

    def get_flat_subdependants(
        self, solved: SolvedDependency[Any]
    ) -> List[DependantProtocol[Any]]:
        """Get an exhaustive list of all of the dependencies of this dependency,
        in no particular order.
        """
        return [dep for group in solved.topsort[1:] for dep in group]

    def _build_task(
        self,
        dependency: DependantProtocol[DependencyType],
        tasks: Dict[DependantProtocol[Any], Tuple[DependantProtocol[Any], Task[Any]]],
        state: ContainerState,
        dag: Dict[
            DependantProtocol[Any],
            Dict[str, DependencyParameter[DependantProtocol[Any]]],
        ],
        parallel: bool,
    ) -> Task[DependencyType]:
        stack = state.stacks[dependency.scope]

        async def bound_call(*args: Any, **kwargs: Any) -> DependencyType:
            assert dependency.call is not None
            if dependency.shared and state.cached_values.contains(dependency.call):
                # use cached value
                res = state.cached_values.get(dependency.call)
            else:
                # if this task is not being parallelized and we are dealing with a sync function
                # then we can just execute it directly
                # otherwise, we wrap it
                if not parallel and not is_coroutine_callable(dependency.call):
                    res = bind_to_stack_as_def_callable(dependency.call, stack=stack)(
                        *args, **kwargs
                    )
                else:
                    res = await bind_to_stack_as_awaitable(
                        dependency.call, stack=stack
                    )(*args, **kwargs)
                if dependency.shared:
                    # caching is allowed, now that we have a value we can save it and start using the cache
                    state.cached_values.set(
                        dependency.call, res, scope=dependency.scope
                    )

            return cast(DependencyType, res)

        task_dependencies: Dict[str, DependencyParameter[Task[DependencyProvider]]] = {}

        for param_name, param in dag[dependency].items():
            task_dependencies[param_name] = DependencyParameter(
                dependency=tasks[param.dependency][1], parameter=param.parameter
            )

        return Task(call=bound_call, dependencies=task_dependencies)

    def _validate_scopes(self, solved: SolvedDependency[Dependency]) -> None:
        """Validate that dependencies all have a valid scope and
        that dependencies only depend on outer scopes or their own scope.
        """
        scopes: Dict[Scope, int] = {
            scope: idx
            for idx, scope in enumerate(reversed(self._state.scopes + [None]))
        }

        def check_is_inner(
            dep: DependantProtocol[Any], subdep: DependantProtocol[Any]
        ) -> None:
            if scopes[dep.scope] > scopes[subdep.scope]:
                raise ScopeViolationError(
                    f"{dep} cannot depend on {subdep} because {subdep}'s"
                    f" scope ({subdep.scope}) is narrower than {dep}'s scope ({dep.scope})"
                )

        def check_scope(dep: DependantProtocol[Any]) -> None:
            if dep.scope not in scopes:
                raise UnknownScopeError(
                    f"Dependency{dep} has an unknown scope {dep.scope}."
                    f" Did you forget to enter the {dep.scope} scope?"
                )

        for dep, params in solved.dag.items():
            check_scope(dep)
            for param in params.values():
                subdep = param.dependency
                check_scope(subdep)
                check_is_inner(dep, subdep)

    async def execute_solved(
        self, solved: SolvedDependency[DependencyType], validate_scopes: bool = True
    ) -> DependencyType:
        """Execute an already solved dependency.

        If you are not dynamically changing scopes, you can run once with `validate_scopes=True`
        and then disable scope validation in subsequent runs with `validate_scope=False`.
        """
        # this mapping uses the hash semantics defined by the implementation of DependantProtocol
        tasks: Dict[
            DependantProtocol[Any], Tuple[DependantProtocol[Any], Task[Any]]
        ] = {}
        async with self.enter_local_scope(None):
            if validate_scopes:
                self._validate_scopes(solved)
            for group in reversed(solved.topsort):
                parallel = len(group) == 1
                for dep in group:
                    if dep not in tasks:
                        tasks[dep] = (
                            dep,
                            self._build_task(
                                dep, tasks, self._state, solved.dag, parallel
                            ),
                        )
            ordered_tasks = [
                [tasks[dep][1] for dep in group] for group in solved.topsort
            ]
            tg: Optional[TaskGroup] = None
            for task_group in reversed(ordered_tasks):
                if len(task_group) > 1:
                    if tg is None:
                        tg = create_task_group()
                    async with tg:
                        for task in task_group:
                            tg.start_soon(task.compute)  # type: ignore
                else:
                    await task_group[0].compute()

        return tasks[solved.dependency][1].get_result()

    async def execute(
        self, dependency: DependantProtocol[DependencyType]
    ) -> DependencyType:
        """Solve and execute a dependency"""
        return await self.execute_solved(self.solve(dependency))



from __future__ import annotations

from typing import Any, Dict, Generic, Hashable, List, TypeVar, Union, overload

from di.dependency import Scope
from di.exceptions import DuplicateScopeError, UnknownScopeError

T = TypeVar("T")
KT = TypeVar("KT", bound=Hashable)
VT = TypeVar("VT")


_unset: Any = object()


class ScopeMap(Generic[KT, VT]):
    """Mapping-like structure to hold binds and cached values.

    The important aspect of this is that we may have thousands of dependencies, but generally only a couple (<10) scopes.
    So this is designed to operate in ~ O(S) time, where S is the number of scopes.
    Imporantly, when we enter or exit a scope (which can happen very frequently, e.g. requests in a web framework),
    no iteration is required: we can simply pop all keys in that scope.

    The interface is *almost* a Mapping[KT, VT] but not quite: some operations like set / __setitem__ require a key, a value *and* a scope.
    We could implement ScopeMap[key, scope] = value, but that can lead to a bit of confusion since we want lookups like ScopMap[key].
    Additionally, operations are not O(1) but rather O(S).
    So we choose to use method names like get() and set() to deal with the API differences and indicate that operations are more expensive than O(1).
    This is also similar to collections.ChainMap, except that we want "named" nodes (Scopes) and ChainMap uses a plain list.
    ChainMap also doesn't allow you to set values anywhere but the left mapping, and we need to set values in arbitrary mappings.
    """

    def __init__(self) -> None:
        self.mappings: Dict[Scope, Dict[KT, VT]] = {}

    def get(self, key: KT) -> VT:
        for mapping in self.mappings.values():
            if key in mapping:
                return mapping[key]
        raise KeyError(key)

    def set(self, key: KT, value: VT, *, scope: Scope) -> None:
        if scope not in self.mappings:
            raise UnknownScopeError(
                f"The scope {scope} is not amongst the known scopes: {self.mappings.keys()}"
            )
        for mapping_scope, mapping in self.mappings.items():
            if scope == mapping_scope:
                mapping[key] = value
            else:
                if key in mapping:
                    mapping.pop(key)

    @overload
    def pop(self, key: KT) -> VT:
        ...

    @overload
    def pop(self, key: KT, default: T) -> Union[VT, T]:
        ...

    def pop(self, key: KT, default: T = _unset) -> Union[VT, T]:
        for mapping in self.mappings.values():
            if key in mapping:
                return mapping.pop(key)
        if default is not _unset:
            return default
        raise KeyError(key)

    def contains(self, key: KT) -> bool:
        for mapping in self.mappings.values():
            if key in mapping:
                return True
        return False

    def get_scope(self, key: KT) -> Scope:
        for scope, mapping in self.mappings.items():
            if key in mapping:
                return scope
        raise KeyError(key)

    def add_scope(self, scope: Scope) -> None:
        if scope in self.mappings:
            raise DuplicateScopeError(f"The scope {scope} already exists!")
        self.mappings[scope] = {}

    def pop_scope(self, scope: Scope) -> None:
        if scope not in self.mappings:
            raise UnknownScopeError(
                f"The scope {scope} is not amongst the known scopes: {self.mappings.keys()}"
            )
        self.mappings.pop(scope)

    def has_scope(self, scope: Scope) -> bool:
        return scope in self.mappings

    def copy(self) -> ScopeMap[KT, VT]:
        new = ScopeMap[KT, VT]()
        new.mappings = self.mappings.copy()
        return new

    def __repr__(self) -> str:  # pragma: no cover, used only for debugging
        values: List[str] = []
        for scope, mapping in self.mappings.items():
            for k, v in mapping.items():
                values.append(f'{repr(k)}: {repr(v)} @ scope="{scope}"')
        return "{" + ", ".join(values) + "}"



from __future__ import annotations

from typing import Any, Awaitable, Callable, Dict, Generic, List, cast

from di._inspect import DependencyParameter
from di.dependency import Dependency, DependencyType

_UNSET = object()


class Task(Generic[DependencyType]):
    def __init__(
        self,
        call: Callable[..., Awaitable[DependencyType]],
        dependencies: Dict[str, DependencyParameter[Task[Dependency]]],
    ) -> None:
        self.call = call
        self.dependencies = dependencies
        self._result: Any = _UNSET

    async def compute(self) -> None:
        positional: List[Task[Dependency]] = []
        keyword: Dict[str, Task[Dependency]] = {}
        for param_name, dep in self.dependencies.items():
            if dep.parameter.kind is dep.parameter.kind.POSITIONAL_ONLY:
                positional.append(dep.dependency.get_result())
            else:
                keyword[param_name] = dep.dependency.get_result()
        self._result = await self.call(*positional, **keyword)

    def get_result(self) -> DependencyType:
        if self._result is _UNSET:
            raise ValueError(
                "`compute()` must be called before `get_result()`; this is likely a bug!"
            )
        return cast(DependencyType, self._result)



from collections import deque
from typing import Deque, Dict, Iterable, List, Mapping, TypeVar

from di.exceptions import CircularDependencyError

T = TypeVar("T")


def topsort(
    root: T,
    dag: Mapping[T, Iterable[T]],
) -> List[List[T]]:
    """Topological sort with grouping."""
    groups: List[List[T]] = []
    dependant_count: Dict[T, int] = dict.fromkeys(dag, 0)
    for subdeps in dag.values():
        for subdep in subdeps:
            dependant_count[subdep] += 1
    q: Deque[T] = deque([root])
    while q:
        group = list(q)
        q.clear()
        for dependant in group:
            for dependency in dag[dependant]:
                try:
                    dependant_count[dependency] -= 1
                except KeyError:
                    raise CircularDependencyError
                if dependant_count[dependency] == 0:
                    dependant_count.pop(dependency)
                    q.append(dependency)
        for dependant in group:
            if dependant_count.pop(dependant, 0) == 0:
                dependant_count.pop(dependant, None)
        groups.append(group)
    for indegree in dependant_count.values():
        if indegree != 0:
            raise CircularDependencyError
    return groups



from __future__ import annotations

import inspect
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Coroutine,
    Dict,
    Generator,
    Hashable,
    Optional,
    Protocol,
    TypeVar,
    Union,
    cast,
    overload,
    runtime_checkable,
)

from di._docstrings import join_docstring_from
from di._inspect import DependencyParameter, get_parameters, infer_call_from_annotation
from di.exceptions import WiringError

DependencyType = TypeVar("DependencyType")

CallableProvider = Callable[..., DependencyType]
CoroutineProvider = Callable[..., Coroutine[Any, Any, DependencyType]]
GeneratorProvider = Callable[..., Generator[DependencyType, None, None]]
AsyncGeneratorProvider = Callable[..., AsyncGenerator[DependencyType, None]]

DependencyProviderType = Union[
    CallableProvider[DependencyType],
    CoroutineProvider[DependencyType],
    GeneratorProvider[DependencyType],
    AsyncGeneratorProvider[DependencyType],
]

Scope = Hashable

Dependency = Any

DependencyProvider = Union[
    AsyncGeneratorProvider[Dependency],
    CoroutineProvider[Dependency],
    GeneratorProvider[Dependency],
    CallableProvider[Dependency],
]


_VARIABLE_PARAMETER_KINDS = (
    inspect.Parameter.VAR_POSITIONAL,
    inspect.Parameter.VAR_KEYWORD,
)


_expected_attributes = ("call", "scope", "shared", "get_dependencies", "is_equivalent")


def _is_dependant_protocol_instance(o: object) -> bool:
    # run cheap attribute checks before running isinstance
    # isinstace is expensive since runs reflection on methods
    # to check argument types, etc.
    for attr in _expected_attributes:
        if not hasattr(o, attr):
            return False
    return isinstance(o, DependantProtocol)


@runtime_checkable
class DependantProtocol(Protocol[DependencyType]):
    """A dependant is an object that can provide the container with:
    - A hash, to compare itself against other dependants
    - A scope
    - A callable that can be used to assemble itself
    - The dependants that correspond to the keyword arguments of that callable
    """

    call: Optional[DependencyProviderType[DependencyType]]
    dependencies: Optional[Dict[str, DependencyParameter[DependantProtocol[Any]]]]
    scope: Scope
    shared: bool

    def __hash__(self) -> int:
        """A unique identifier for this dependency.

        By default, dependencies are identified by their call attribute.
        This can be overriden to introduce other semantics, e.g. to involve the scope or custom attrbiutes
        in dependency identification.
        """
        raise NotImplementedError

    def __eq__(self, o: object) -> bool:
        """Used in conjunction with __hash__ for mapping lookups of dependencies.
        Generally, this should have the same semantics as __hash__ but can check for object identity.
        """
        raise NotImplementedError

    def is_equivalent(self, other: DependantProtocol[Any]) -> bool:
        """Copare two DependantProtocol implementers for equality.

        By default, the two are equal only if they share the same callable and scope.

        If two dependencies share the same hash but are not equal, the Container will
        report an error.
        This is commonly caused by using the same callable under two different scopes.
        To remedy this, you can either wrap the callable to give it two different hashes/ids,
        or you can create a DependantProtocol implementation that overrides __hash__ and/or __eq__.
        """
        raise NotImplementedError

    def get_dependencies(
        self,
    ) -> Dict[str, DependencyParameter[DependantProtocol[Any]]]:
        """A cache on top of `gather_dependencies()`"""
        raise NotImplementedError

    def gather_parameters(self) -> Dict[str, inspect.Parameter]:
        """Collect parameters that this dependency needs to construct itself.

        Generally, this means introspecting into this dependencies own callable.
        """
        raise NotImplementedError

    def create_sub_dependant(
        self, call: DependencyProvider, scope: Scope, shared: bool
    ) -> DependantProtocol[Any]:
        """Create a Dependant instance from a sub-dependency of this Dependency.

        This is used in the scenario where a transient dependency is inferred from a type hint.
        For example:
        >>> class Foo:
        >>>     ...
        >>> def foo_factory(foo: Foo) -> Foo:
        >>>     return foo
        >>> def parent(foo: Dependency(foo_factory)):
        >>>    ...
        In this scenario, `Dependency(foo_factory)` will call `create_sub_dependant(Foo)`.
        """
        raise NotImplementedError

    def infer_call_from_annotation(
        self, param: inspect.Parameter
    ) -> DependencyProvider:
        """Called when the dependency was not explicitly passed a callable.

        It is important to note that param in this context refers to the parameter in this
        Dependant's parent.
        For example, in the case of `def func(thing: Something = Dependency())` this method
        will be called with a Parameter corresponding to Something.
        """
        raise NotImplementedError


class Dependant(DependantProtocol[DependencyType]):
    @overload
    def __init__(
        self,
        call: Optional[AsyncGeneratorProvider[DependencyType]] = None,
        scope: Optional[Scope] = None,
        shared: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[CoroutineProvider[DependencyType]] = None,
        scope: Optional[Scope] = None,
        shared: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[GeneratorProvider[DependencyType]] = None,
        scope: Optional[Scope] = None,
        shared: bool = True,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        call: Optional[CallableProvider[DependencyType]] = None,
        scope: Optional[Scope] = None,
        shared: bool = True,
    ) -> None:
        ...

    def __init__(
        self,
        call: Optional[DependencyProviderType[DependencyType]] = None,
        scope: Scope = None,
        shared: bool = True,
    ) -> None:
        self.call = call
        self.scope = scope
        self.dependencies: Optional[
            Dict[str, DependencyParameter[DependantProtocol[Any]]]
        ] = None
        self.shared = shared

    @join_docstring_from(DependantProtocol[Any].create_sub_dependant)
    def create_sub_dependant(
        self, call: DependencyProvider, scope: Scope, shared: bool
    ) -> DependantProtocol[Any]:
        return Dependant[Any](call=call, scope=scope, shared=shared)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(call={self.call}, scope={self.scope})"

    @join_docstring_from(DependantProtocol[Any].__hash__)
    def __hash__(self) -> int:
        return hash(self.call)

    @join_docstring_from(DependantProtocol[Any].__eq__)
    def __eq__(self, o: object) -> bool:
        if type(self) != type(o):
            return False
        assert isinstance(o, type(self))
        return self.call is o.call

    @join_docstring_from(DependantProtocol[Any].is_equivalent)
    def is_equivalent(self, other: DependantProtocol[Any]) -> bool:
        return self.call is other.call and other.scope == self.scope

    @join_docstring_from(DependantProtocol[Any].get_dependencies)
    def get_dependencies(
        self,
    ) -> Dict[str, DependencyParameter[DependantProtocol[Any]]]:
        if self.dependencies is None:  # type: ignore
            self.dependencies = self.gather_dependencies()
        return self.dependencies

    @join_docstring_from(DependantProtocol[Any].gather_parameters)
    def gather_parameters(self) -> Dict[str, inspect.Parameter]:
        assert self.call is not None, "Cannot gather parameters without a bound call"
        return get_parameters(self.call)

    @join_docstring_from(DependantProtocol[Any].infer_call_from_annotation)
    def infer_call_from_annotation(
        self, param: inspect.Parameter
    ) -> DependencyProvider:
        if param.annotation is param.empty:
            raise WiringError(
                "Cannot wire a parameter with no default and no type annotation"
            )
        return infer_call_from_annotation(param)

    def gather_dependencies(
        self,
    ) -> Dict[str, DependencyParameter[DependantProtocol[Any]]]:
        """Collect this dependencies sub dependencies.

        The returned dict corresponds to keyword arguments that will be passed
        to this dependencies `call` after all sub-dependencies are themselves resolved.
        """
        assert (
            self.call is not None
        ), "Container should have assigned call; this is a bug!"
        res: Dict[str, DependencyParameter[DependantProtocol[Any]]] = {}
        for param_name, param in self.gather_parameters().items():
            if param.kind in _VARIABLE_PARAMETER_KINDS:
                raise WiringError(
                    "Dependencies may not use variable positional or keyword arguments"
                )
            if _is_dependant_protocol_instance(param.default):
                sub_dependant = cast(DependantProtocol[Any], param.default)
                if sub_dependant.call is None:
                    sub_dependant.call = sub_dependant.infer_call_from_annotation(param)
            elif param.default is param.empty:
                sub_dependant = self.create_sub_dependant(
                    call=self.infer_call_from_annotation(param),
                    scope=self.scope,
                    shared=self.shared,
                )
            else:
                continue  # pragma: no cover
            res[param_name] = DependencyParameter(
                dependency=sub_dependant, parameter=param
            )
        return res



class WiringError(Exception):
    pass


class UnknownScopeError(Exception):
    pass


class DuplicateScopeError(Exception):
    pass


class DuplicatedDependencyError(Exception):
    pass


class CircularDependencyError(Exception):
    pass


class ScopeConflictError(Exception):
    pass


class ScopeViolationError(Exception):
    pass


class DependencyRegistryError(Exception):
    pass



from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from typing import Any, AsyncGenerator, ContextManager, Dict, Generator, List, Optional

from di._scope_map import ScopeMap
from di.dependency import (
    DependantProtocol,
    DependencyProvider,
    DependencyProviderType,
    DependencyType,
    Scope,
)
from di.exceptions import DuplicateScopeError


class ContainerState:
    def __init__(self) -> None:
        self.binds = ScopeMap[DependencyProvider, DependantProtocol[Any]]()
        self.cached_values = ScopeMap[DependencyProvider, Any]()
        self.stacks: Dict[Scope, AsyncExitStack] = {}

    def copy(self) -> "ContainerState":
        new = ContainerState()
        new.binds = self.binds.copy()
        new.stacks = self.stacks.copy()
        new.cached_values = self.cached_values.copy()
        return new

    @asynccontextmanager
    async def enter_scope(self, scope: Scope) -> AsyncGenerator[None, None]:
        if scope in self.stacks:
            raise DuplicateScopeError(f"Scope {scope} has already been entered!")
        stack = AsyncExitStack()
        self.stacks[scope] = stack
        if not self.binds.has_scope(scope):
            self.binds.add_scope(scope)
        self.cached_values.add_scope(scope)
        try:
            yield
        finally:
            await stack.aclose()
            self.stacks.pop(scope)
            self.cached_values.pop_scope(scope)

    @property
    def scopes(self) -> List[Scope]:
        return list(self.stacks.keys())

    def bind(
        self,
        provider: DependantProtocol[DependencyType],
        dependency: DependencyProviderType[DependencyType],
        scope: Scope,
    ) -> ContextManager[None]:
        if not self.binds.has_scope(scope):
            self.binds.add_scope(scope)
        previous_scope: Optional[Scope]
        previous_provider: Optional[DependantProtocol[Any]]
        try:
            previous_provider = self.binds.get(dependency)
            previous_scope = self.binds.get_scope(dependency)
        except KeyError:
            previous_provider = None
            previous_scope = None

        self.binds.set(dependency, provider, scope=scope)
        if self.cached_values.contains(dependency):
            self.cached_values.pop(dependency)

        @contextmanager
        def unbind() -> Generator[None, None, None]:
            try:
                yield
            finally:
                self.binds.pop(dependency)
                if previous_provider is not None:
                    self.binds.set(dependency, previous_provider, scope=previous_scope)

        return unbind()
